#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
In this module we define the class :class:`Ph_sel` used to specify a
"selection" of a sub-set of photons/timestamps  (i.e. all-photons,
Donor-excitation-period photons, etc...).

A photon selection is one of the base *photon streams* or a combination of
them. Base *photon streams* are photon from the donor (or acceptor) emission
channel detected during the donor (or acceptor) excitation period. For
non-ALEX data there is only the donor excitation period.

The following table shows base *photon streams* for smFRET data (non-ALEX):

=================   =====================
Photon selection    Syntax
=================   =====================
D-emission          `Ph_sel(Dex='Dem')`
A-emission          `Ph_sel(Dex='Aem')`
=================   =====================

and for ALEX data:

==============================   =====================
Photon selection                 Syntax
==============================   =====================
D-emission during D-excitation   `Ph_sel(Dex='Dem')`
A-emission during D-excitation   `Ph_sel(Dex='Aem')`
D-emission during A-excitation   `Ph_sel(Aex='Dem')`
A-emission during A-excitation   `Ph_sel(Aex='Aem')`
==============================   =====================

Additionally, all the photons can be selected with `Ph_sel('all')` (that is a
shortcut for  `Ph_sel(Dex='DAem', Aex='DAem')`.

Examples:

    - `Ph_sel(Dex='DAem', Aex='DAem')` or `Ph_sel('all')` select all photons.
    - `Ph_sel(Dex='DAem')` selects only donor and acceptor photons
      emitted during donor excitation. These are all the photons for
      non-ALEX data.
    - `Ph_sel(Dex='Aem', Aex='Aem')` selects all the photons detected from
      the acceptor-emission channel.

The documentation for the :class:`Ph_sel` class follows.

"""

from collections import namedtuple


# Implementation Rationale:
#
#   This class is implemented as a named tuple that is an immutable object.
#   This means the Ph_sel objects can be used as dictionary keys. This would
#   not be possible if Ph_sel were a dict. The __new__ method is just a nicety
#   to allow only valid arguments values and to throw meaningful exceptions
#   otherwise.
#
class Ph_sel(namedtuple('Ph_sel', ['Dex', 'Aex'])):
    """Class that describes a selection of photons.

    This class takes two arguments `Dex` and `Aex`.
    Valid values for the arguments are the strings 'DAem', 'Dem', 'Aem' or
    `None`. These values select, respectively, donor+acceptor, donor-only,
    acceptor-only or no photons during an excitation period (`Dex` or `Aex`).

    The class must be called with at least one keyword argument or using
    the string 'all' as the only argument. Calling `Ph_sel('all')` is
    equivalent to `Ph_sel(Dex='DAem', Aex='DAem')`.
    Not specifying a keyword argument is equivalent to setting it to None.

    """
    valid_values = ('DAem', 'Dem', 'Aem', None)
    def __new__(cls, Dex=None, Aex=None):
        if Dex is None and Aex is None:
            raise ValueError("You need to specify at least one argument "
                             "(Dex, Aex or 'all').")
        if Dex == 'all':
            return super(Ph_sel, cls).__new__(cls, 'DAem', 'DAem')
        if Dex not in cls.valid_values or Aex not in cls.valid_values:
            raise ValueError("Invalid value %s. Valid values are "
                             "'DAem', 'Dem' or 'Aem' (or None)." %\
                             str((Dex, Aex)))
        return super(Ph_sel, cls).__new__(cls, Dex, Aex)

    def __str__(self):
        labels = {Ph_sel('all'): 'all',
                  Ph_sel(Dex='Dem'): 'DexDem', Ph_sel(Dex='Aem'): 'DexAem',
                  Ph_sel(Aex='Aem'): 'AexAem', Ph_sel(Aex='Dem'): 'AexDem',
                  Ph_sel(Dex='DAem'): 'Dex', Ph_sel(Aex='DAem'): 'Aex',
                  Ph_sel(Dex='DAem', Aex='Aem'): 'DexDAem_AexAem'}
        return labels.get(self, repr(self))



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module contains functions for direct fitting of burst populations
(FRET peaks) without passing through a FRET histogram.

This module provides a standard interface for different fitting algorithms.
"""

from __future__ import print_function, absolute_import
from builtins import range, zip

import numpy as np
from scipy.stats import binom, expon
from scipy.optimize import minimize_scalar, leastsq

from .fit import gaussian_fitting as gf


def sim_nd_na(E,N=1000, size_mean=100):
    """Simulate an exponential-size burst distribution with binomial (nd,na)
    """
    nt = np.ceil(expon.rvs(scale=size_mean, size=N)).astype(int)
    na = binom.rvs(nt, E)
    nd = nt - na
    return nd, na


def log_likelihood_binom(E, nd, na):
    """Likelihood function for (nd,na) to be from a binom with p=E (no BG)."""
    return -(np.log(binom.pmf(na, na+nd, E))).sum()

def fit_E_binom(nd, na, noprint=False, method='c', **kwargs):
    """Fit the E with MLE using binomial distribution.
    method  ('a','b', or 'c') choose how to handle negative (nd,na) values.
    """
    assert method in ['a', 'b', 'c']
    nd, na = np.round(nd).astype(int), np.round(na).astype(int)

    # The binomial distribution can not handle negative values
    # so we must find a way to "remove" the negativa values
    # a. remove bursts with neg. values, but we can skew the fit
    # b. remove bursts with nd < nd[na<0].max(), but few bursts left
    # c. remove bursts with na+nd < max(nd[na<0].max(),na[nd<0].max())
    #    giving a bit more bursts
    # NOTE: b and c have corner cases in which there are neg. bursts left
    pos_bursts = (nd>=0)*(na>=0)
    if (-pos_bursts).any():
        if not noprint: print("WARNING: Discarding negative burst sizes.")
        if method == 'a':
            # a. remove bursts with neg. values
            nd, na = nd[pos_bursts], na[pos_bursts]
        elif method == 'b':
            # b. Cut all the part with na<0 to have a less skewed estimation
            if (na < 0).any():
                nd_min = nd[na<0].max()
                nd, na = nd[nd>nd_min], na[nd>nd_min]
        elif method == 'c':
            # c. remove bursts with na+nd < max(nd[na<0].max(),na[nd<0].max())
            nt_min = 0
            if (na < 0).any():
                nt_min = nd[na<0].max()
            if (nd < 0).any():
                nt_min = max([na[nd<0].max(), nt_min])
            nd, na = nd[nd+na>nt_min], na[nd+na>nt_min]

        if not noprint: print(" - Bursts left:", nd.size)
        assert (nd>=0).all() and (na>=0).all()
    min_kwargs = dict(bounds=(0,1), method='bounded',
            options={'disp':1, 'xtol': 1e-6})
    min_kwargs.update(**kwargs)
    res = minimize_scalar(log_likelihood_binom, args=(nd,na), **min_kwargs)
    return res.x


def log_likelihood_poisson_nt(E, nd, na, bg_a):
    """Likelihood function for na extracted from Poisson. nd, na BG corrected.
    """
    k = 1.*E/(1-E)
    logP = 0
    nt = nd + na
    for nti, nai, bg_ai in zip(nt, na, bg_a):
        lam_ai = k*(nti)/(1+k) + bg_ai
        logP += (nai+bg_ai)*np.log(lam_ai) - lam_ai
    return -logP

def log_likelihood_poisson_na(E, nd, na, bg_a):
    """Likelihood function for na extracted from Poisson. nd, na BG corrected.
    """
    k = 1.*E/(1-E)
    logP = 0
    for ndi, nai, bg_ai in zip(nd, na, bg_a):
        lam_ai = k*ndi + bg_ai
        logP += (nai+bg_ai)*np.log(lam_ai) - lam_ai
    return -logP

def log_likelihood_poisson_nd(E, nd, na, bg_d):
    """Likelihood function for nd extracted from Poisson. nd, na BG corrected.
    """
    k = 1.*E/(1-E)
    logP = 0
    for ndi, nai, bg_di in zip(nd, na, bg_d):
        lam_di = (1./k)*nai + bg_di
        logP += (ndi+bg_di)*np.log(lam_di) - lam_di
    return -logP

def fit_E_poisson_nt(nd, na, bg_a, **kwargs):
    """Fit the E using MLE with na extracted from a Poisson.
    """
    min_kwargs = dict(bounds=(-0.1,1.1), method='bounded',
            options={'disp':1, 'xtol': 1e-6})
    min_kwargs.update(**kwargs)
    res = minimize_scalar(log_likelihood_poisson_nt, args=(nd, na, bg_a),
            **min_kwargs)
    E = res.x
    return E

def fit_E_poisson_na(nd, na, bg_a, **kwargs):
    """Fit the E using MLE with na extracted from a Poisson.
    """
    min_kwargs = dict(bounds=(-0.1,1.1), method='bounded',
            options={'disp':1, 'xtol': 1e-6})
    min_kwargs.update(**kwargs)
    res = minimize_scalar(log_likelihood_poisson_na, args=(nd, na, bg_a),
            **min_kwargs)
    E = res.x
    return E

def fit_E_poisson_nd(nd, na, bg_d, **kwargs):
    """Fit the E using MLE with nd extracted from a Poisson.
    """
    min_kwargs = dict(bounds=(-0.1,1.1), method='bounded',
            options={'disp':1, 'xtol': 1e-6})
    min_kwargs.update(**kwargs)
    res = minimize_scalar(log_likelihood_poisson_nd, args=(nd, na, bg_d),
            **min_kwargs)
    E = res.x
    return E

def fit_E_hist(nd, na, gamma=1., **kwargs):
    """Fit E using the histogram curve-fit (see gaussian_fit_hist).
    You can specify `weights` that will be passed to the `histogram` function.
    """
    E = na/(na + gamma*nd)
    if 'weights' in kwargs:
        w = get_weights(nd, na, weights=kwargs['weights'])
        kwargs.update(weights=w)
    E_mu, E_sig = gf.gaussian_fit_hist(E, mu0=0.5, sigma0=0.1, **kwargs)
    return E_mu

def fit_E_cdf(nd, na, gamma=1., **kwargs):
    """Fit E using the CDF curve-fit (see gaussian_fit_cdf).
    No weights are possible with this method.
    """
    E = na/(na + gamma*nd)
    E_mu, E_sig = gf.gaussian_fit_cdf(E, mu0=0.5, sigma0=0.1, **kwargs)
    return E_mu

def get_dist_euclid(nd, na, E_fit=None, slope=None):
    """Returns the euclidean distance of (nd,na) from a fit line.
    The fit line is specified by `slope` or by `E_fit`. Intercept is always 0.
    """
    if slope is None:
        try:
            slope = E_fit/(1-E_fit)
        except TypeError:
            # this happens if E_fit is None
            raise ValueError('One among `E_fit` or `slope` must be not None.')

    # For each burst (nd,na) find the closest point on the fitted line
    nd_l = (slope*na + nd)/(slope**2+1)
    na_l = slope*nd_l
    distances = np.sqrt((nd-nd_l)**2 + (na-na_l)**2)
    return distances

def get_weights(nd, na, weights, naa=0, gamma=1., widths=None):
    """Return burst weigths computed according to different criteria.

    The burst size is computed as `nd*gamma + na + naa`.

    Arguments:
        nd, na, naa (1D arrays): photon counts in each burst.
        gamma (float): gamma factor used for corrected burst size.
        width (None array): array of burst durations used when
            weights='brightness'
        weights (string or None): type of weights, possible weights are:
            'size' burst size, 'size_min' burst size - min(burst size),
            'size2' (burst size)^2, 'sqrt' sqrt(burst size),
            'inv_size' 1/(burst size), 'inv_sqrt' 1/sqrt(burst size),
            'cum_size' CDF_of_burst_sizes(burst size),
            'cum_size2' CDF_of_burst_sizes(burst size)^2,
            'brightness' the burst size divided by the burst width.
            If None returns uniform weights.
        widths (1D array): bursts duration in seconds, needed only when
            weights = 'brightness'.

    Returns:
        1D array of weigths, one element per burst.
    """
    nt = nd*gamma + na + naa
    if weights is None:           # all weights the same
        weights = np.ones(nd.size)
    elif weights == 'size':       # weight = burst size (with gamma)
        weights = nt
    elif weights == 'size_min':   # weight = burst size - min(burst size)
        weights = nt - nt.min() + 1
    elif weights == 'size2':      # weight = (burst size)^2
        weights = nt**2
    elif weights == 'sqrt':       # weigth = sqrt(burst size)
        weights = np.sqrt(nt)
    elif weights == 'inv_size':   # weight = 1/(burst size)
        weights = 1./(nt)
    elif weights == 'inv_sqrt':   # weight = 1/sqrt(burst size)
        weights = 1./np.sqrt(nt)
    elif weights == 'cum_size':   # weight = CDF_of_burst_sizes(burst size)
        ecdf = [np.sort(nt), 1.*np.arange(1, nt.size+1)/nt.size]
        weights = np.interp(nt, ecdf[0], ecdf[1], left=0, right=1)
    elif weights == 'cum_size2':   # weight = CDF_of_burst_sizes(burst size)^2
        ecdf = [np.sort(nt), 1.*np.arange(1, nt.size+1)/nt.size]
        weights = np.interp(nt, ecdf[0], ecdf[1]**2, left=0, right=1)
    elif weights == 'brightness':
        assert widths is not None, \
            "The widths argument is needed to compute the 'brigness' weights"
        weights = nt/widths
    else:
        raise ValueError("Wrong weights name.")
    assert weights.size == nd.size
    return weights


def fit_E_slope(nd, na, weights=None, gamma=1.):
    """Fit E with a least-squares fitting of slope on (nd,na) plane."""
    weights = get_weights(nd, na, weights=weights, gamma=gamma)
    #err_fun = lambda k, x, y, w: (y - k*x)*w
    err_fun = lambda k, x, y, w: w*get_dist_euclid(nd=x, na=y, slope=k)
    res = leastsq(err_fun, x0=1, args=(nd, na, weights))
    #print(res)
    return res[0][0]/(res[0][0]+1)

def fit_E_E_size(nd, na, weights=None, gamma=1., gamma_correct=False):
    """Fit the E with least-square minimization of errors on burst E values."""
    weights = get_weights(nd, na, weights=weights, gamma=gamma)
    if gamma_correct:
        err_fun = lambda E_fit, d, a, w: (E_fit - 1.*a/(gamma*d+a))*w
    else:
        err_fun = lambda E_fit, d, a, w: (E_fit - 1.*a/(d+a))*w
    res = leastsq(err_fun, x0=0.5, args=(nd, na, weights))
    return res[0][0]

def fit_E_m(nd, na, weights=None, gamma=1., gamma_correct=False):
    """Fit the E with a weighted mean of burst E values."""
    weights = get_weights(nd, na, weights=weights, gamma=gamma)
    if gamma_correct:
        E = na/(na + gamma*nd)
    else:
        E = na/(na + nd)
    E_fit = np.dot(E, weights)/weights.sum()
    return E_fit



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module provides functions to store and load background fit data
to and from an HDF5 file.

The functions here assume to find an open pyTables file reference in
the :class:`Data` attribute `.bg_data_file`.
"""

from __future__ import  absolute_import
from builtins import range, zip

import os
import numpy as np

from .utils.misc import pprint
from .ph_sel import Ph_sel


def remove_cache(dx):
    """Remove all the saved background data."""
    assert 'bg_data_file' in dx
    pprint(' * Removing all the cached background data ... ')
    if '/background' in dx.bg_data_file:
        dx.bg_data_file.remove_node('/background', recursive=True)
        dx.bg_data_file.flush()
    pprint('[DONE]\n')

def _get_bg_arrays_info(dx):
    """Return a dict of background numeric data description.

    The keys of the returned dict are names and the values are a string
    description of numeric data related to the background fit.
    This info is used to set the title attribute in the HDF5 arrays.
    """
    bg_arrays_info = dict(
        bg = 'BG rate in each CH vs time (Total)',
        bg_dd = 'BG rate in each CH vs time (D_em during D_ex)',
        bg_ad = 'BG rate in each CH vs time (A_em during D_ex)',
        bg_da = 'BG rate in each CH vs time (D_em during A_ex)',
        bg_aa = 'BG rate in each CH vs time (A_em during A_ex)',
        Lim = 'Index of first and last timestamp in each period',
        Ph_p = 'First and last timestamp in each period',
        nperiods = 'Number of time periods in which BG is computed',
        bg_time_s = 'Time duration of the period (windows in which computing BG)',
        bg_auto_th = '1 if the bg threshold was computed automatically, else 0'
    )

    bg_arrays_info_auto = dict(
        bg_auto_th_us0 = 'Threshold used for the initial bg rate estimation.',
        bg_auto_F_bg = 'Factor that multiplies the initial rate estimation to '
                       'get the auto threshold',
    )
    bg_arrays_info_noauto = dict(
        bg_th_us_user = ('Waiting time thresholds for BG fit. '
                         'This array contains 5 thresholds for different '
                         'photon selections. In the order: all photons, '
                         ' D_em-D_ex, A_em-D_ex, D_em-A_ex, A_em-A_ex.')
    )

    if dx.bg_auto_th:
        bg_arrays_info.update(**bg_arrays_info_auto)
    else:
        bg_arrays_info.update(**bg_arrays_info_noauto)
    return bg_arrays_info

def bg_save_hdf5(dx):
    """Save background data to HDF5 file."""
    assert 'bg_data_file' in dx
    assert 'bg' in dx

    bg_arrays_info = _get_bg_arrays_info(dx)
    bg_attr_names = ['bg_fun', 'bg_fun_name']
    if '/background' not in dx.bg_data_file:
        dx.bg_data_file.create_group('/', 'background',
                                  title='Background estimation data')
    ## Save the bg data
    group_name = _get_bg_groupname(dx)
    if group_name in dx.bg_data_file:
        dx.bg_data_file.remove_node(group_name, recursive=True)

    bg_group = dx.bg_data_file.create_group(os.path.dirname(group_name),
                                         os.path.basename(group_name),
                                         createparents=True)
    # Save arrays and scalars
    pprint('\n - Saving arrays/scalars: ')
    for name, info in bg_arrays_info.items():
        pprint(name + ', ')
        arr = np.array(dx[name])
        dx.bg_data_file.create_array(bg_group, name, obj=arr, title=info)

    # Save the attributes
    pprint('\n - Saving HDF5 attributes: ')
    for attr in bg_attr_names:
        pprint(attr + ', ')
        bg_group._v_attrs[attr] = dx[attr]
    pprint('\n')
    dx.bg_data_file.flush()

def bg_load_hdf5(dx, group_name):
    """Load background data from a HDF5 file."""
    assert 'bg_data_file' in dx
    if group_name not in dx.bg_data_file:
        pprint('Group "%s" not found in the HDF5 file.' % group_name)
        return

    ## Load the bg data
    bg_arrays = dict()
    bg_attrs = dict()

    bg_group = dx.bg_data_file.get_node(group_name)

    # Load arrays and scalars
    pprint('\n - Loading arrays/scalars: ')
    for node in bg_group._f_list_nodes():
        name = node.name
        pprint(name + ', ')
        #title = node.title
        arr = bg_group._f_get_child(name)
        bg_arrays[name] = arr.read()
    dx.add(**bg_arrays)

    # Load the attributes
    pprint('\n - Loading HDF5 attributes: ')
    for attr in bg_group._v_attrs._f_list():
        pprint(attr + ', ')
        bg_attrs[attr] = bg_group._v_attrs[attr]
    dx.add(**bg_attrs)
    pprint('\n')

    if 'bg_da' not in dx:
        pprint('   WARNING: Adding missing bg_da as zeros.\n')
        dx.add(bg_da=[np.zeros(bg.shape) for bg in dx.bg_dd])
    if 'bg_ph_sel' not in dx:
        pprint('   WARNING: Adding missing bg_ph_sel.\n')
        dx.add(bg_ph_sel=Ph_sel('all'))

    pprint(' - Generating additional fields: ')
    in_map = ['', '_dd', '_ad', '_da', '_aa']
    out_map = ['_m', '_dd', '_ad', '_da', '_aa']
    new_attrs = {}
    for in_s, out_s in zip(in_map, out_map):
        assert 'bg' + in_s in dx
        pprint('rate' + out_s + ', ')
        new_attrs['rate' + out_s] = [bg.mean() for bg in dx['bg' + in_s]]
    dx.add(**new_attrs)
    pprint('\n')

def _get_bg_groupname(dx, time_s=None):
    """Get the HDF5 group name for the background data.

    Arguments:
        dx (Data): object containing the measurement data
        time_s (None or float): if None the value is taken from dx.bg_time_s
    Returns:
        A string for the background group name.
    """
    if time_s is None and 'bg' not in dx:
        pprint('You need to compute the background or provide time_s.')
        return
    time_slice = time_s if time_s is not None else dx.bg_time_s
    return '/background/time_%ds' % time_slice

def _bg_is_cached(dx, signature):
    """Returns wheter background with given `signature` is in disk the cache.
    """
    if 'bg_data_file' in dx:
        bg_groupname = _get_bg_groupname(dx, time_s=signature['time_s'])
        if bg_groupname in dx.bg_data_file:
            # At least we have a group with the right time_s
            # Check whether its signature matches the current one
            group = dx.bg_data_file.get_node(bg_groupname)
            if signature == group._v_attrs.signature:
                return True
    return False

def _bg_add_signature(dx, signature):
    """Add the signature attr to current background group.
    """
    assert 'bg_data_file' in dx
    bg_groupname = _get_bg_groupname(dx, time_s=signature['time_s'])
    assert bg_groupname in dx.bg_data_file

    group = dx.bg_data_file.get_node(bg_groupname)
    group._v_attrs.signature = signature
    dx.bg_data_file.flush()

def calc_bg_cache(dx, fun, time_s=60, tail_min_us=500, F_bg=2, recompute=False):
    """Cached version of `.calc_bg()` method."""

    curr_call_signature = dict(fun_name=fun.__name__, time_s=time_s,
                               tail_min_us=tail_min_us, F_bg=F_bg)
    if _bg_is_cached(dx, curr_call_signature) and not recompute:
        # Background found in cache. Load it.
        pprint(' * Loading BG rates from cache ... ')
        bg_groupname = _get_bg_groupname(dx, time_s=time_s)
        dx._clean_bg_data()
        bg_load_hdf5(dx, bg_groupname)
        pprint(' [DONE]\n')
    else:
        # Background not found in cache. Compute it.
        pprint(' * No cached BG rates, recomputing:\n')
        dx.calc_bg(fun=fun, time_s=time_s, tail_min_us=tail_min_us, F_bg=F_bg)
        if 'bg_data_file' in dx:
            pprint(' * Storing BG  to disk ... ')
            # And now store it to disk
            bg_save_hdf5(dx)
            _bg_add_signature(dx, curr_call_signature)
            pprint(' [DONE]\n')


def test_calc_bg_cache(dx):
    pass




# encoding: utf-8
#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
The `fretmath` module contains functions to compute corrected FRET efficiency
from the proximity ratio and vice-versa.

For derivation see notebook: "Derivation of FRET and S correction formulas.ipynb"
(`link <http://nbviewer.jupyter.org/github/tritemio/notebooks/blob/master/Derivation%20of%20FRET%20and%20S%20correction%20formulas.ipynb>`__).

"""

from __future__ import division, print_function
from builtins import range, zip

import numpy as np

##
# Comple corrections
#
def correct_E_gamma_leak_dir(Eraw, gamma=1, leakage=0, dir_ex_t=0):
    """Compute corrected FRET efficency from proximity ratio `Eraw`.

    For the inverse function see :func:`uncorrect_E_gamma_leak_dir`.

    Arguments:
        Eraw (float or array): proximity ratio (only background correction,
            no gamma, leakage or direct excitation)
        gamma (float): gamma factor
        leakage (float): leakage coefficient
        dir_ex_t (float): coefficient expressing the direct excitation as
            n_dir = dir_ex_t * (na + gamma*nd). In terms of physical
            parameters it is the ratio of acceptor over donor absorption
            cross-sections at the donor-excitation wavelength.

    Returns
        Corrected FRET effciency
    """
    if isinstance(Eraw, (list, tuple)):
        Eraw = np.asarray(Eraw)
    return (Eraw*(leakage + dir_ex_t*gamma + 1) - leakage - dir_ex_t*gamma) \
           / (Eraw*(leakage -  gamma + 1) - leakage + gamma)

def uncorrect_E_gamma_leak_dir(E, gamma=1, leakage=0, dir_ex_t=0):
    """Compute proximity ratio from corrected FRET efficiency `E`.

    This function is the inverse of :func:`correct_E_gamma_leak_dir`.

    Arguments:
        E (float or array): corrected FRET efficiency
        gamma (float): gamma factor
        leakage (float): leakage coefficient
        dir_ex_t (float): direct excitation coefficient expressed as
            n_dir = dir_ex_t * (na + gamma*nd). In terms of physical
            parameters it is the ratio of absorption cross-section at
            donor-excitation wavelengths of acceptor over donor.

    Returns
        Proximity ratio (reverses gamma, leakage and direct excitation)
    """
    if isinstance(E, (list, tuple)):
        E = np.asarray(E)
    return (E*(gamma - leakage) + leakage + dir_ex_t*gamma) \
           / (E*(gamma - leakage - 1) + leakage + dir_ex_t*gamma + 1)


##
# Single parameter correction
#
# NOTE: Applying the single parameter correction one after the other
#       IS NOT equivalent to applying the comple correction in one step.
#       In fact chaining the single correction gives mathematically the wrong
#       result, even if it may be in some cases close enough to the
#       mathematically accurate value.
#
def gamma_correct_E(Eraw, gamma):
    """Apply gamma correction to the uncorrected FRET `Eraw`.

    For the inverse see :func:`gamma_uncorrect_E`.
    """
    if isinstance(Eraw, (list, tuple)):
        Eraw = np.asarray(Eraw)
    return Eraw / (gamma - gamma*Eraw + Eraw)

def gamma_uncorrect_E(E, gamma):
    """Reverse gamma correction and return uncorrected FRET.

    For the inverse see :func:`gamma_correct_E`.
    """
    if isinstance(E, (list, tuple)):
        E = np.asarray(E)
    return gamma*E/(1 - E + gamma*E)


def leakage_correct_E(Eraw, leakage):
    """Apply leakage correction to the uncorrected FRET `Eraw`.

    For the inverse see :func:`leakage_uncorrect_E`.
    """
    if isinstance(Eraw, (list, tuple)):
        Eraw = np.asarray(Eraw)
    return (Eraw*(leakage + 1) - leakage) / (Eraw*leakage - leakage + 1)

def leakage_uncorrect_E(E, leakage):
    """Reverse leakage correction and return uncorrected FRET.

    For the inverse see :func:`leakage_correct_E`.
    """
    if isinstance(E, (list, tuple)):
        E = np.asarray(E)
    return (E + leakage - E*leakage) / (1 + leakage - E*leakage)


def dir_ex_correct_E(Eraw, dir_ex_t):
    """Apply direct excitation correction to the uncorrected FRET `Eraw`.

    The coefficient `dir_ex_t` expresses the direct excitation as
    n_dir = dir_ex_t * (na + gamma*nd). In terms of physical
    parameters it is the ratio of acceptor over donor absorption
    cross-sections at the donor-excitation wavelength.

    For the inverse see :func:`dir_ex_uncorrect_E`.
    """
    if isinstance(Eraw, (list, tuple)):
        Eraw = np.asarray(Eraw)
    return Eraw*(dir_ex_t + 1) - dir_ex_t

def dir_ex_uncorrect_E(E, dir_ex_t):
    """Reverse direct excitation correction and return uncorrected FRET.

    For the inverse see :func:`dir_ex_correct_E`.
    """
    if isinstance(E, (list, tuple)):
        E = np.asarray(E)
    return (E + dir_ex_t) / (dir_ex_t + 1)

def correct_S(Eraw, Sraw, gamma, leakage, dir_ex_t):
    """Correct S values for gamma, leakage and direct excitation.

    Arguments:
        Eraw (scalar or array): uncorrected ("raw") E after only background
            correction (no gamma, leakage or direct excitation).
        Sraw (scalar or array): uncorrected ("raw") S after only background
            correction (no gamma, leakage or direct excitation).
        gamma (float): gamma factor.
        leakage (float): donor emission leakage into the acceptor channel.
        dir_ex_t (float): direct acceptor excitation by donor laser.
            Defined as ``n_dir = dir_ex_t * (na + g nd)``. The dir_ex_t
            coefficient is the ratio between D and A absorbtion cross-sections
            at the donor-excitation wavelength.

    Returns
        Corrected S (stoichiometry), same size as `Sraw`.
    """
    if isinstance(Eraw, (list, tuple)):
        Eraw = np.asarray(Eraw)
    if isinstance(Sraw, (list, tuple)):
        Sraw = np.asarray(Sraw)
    return (Sraw*(Eraw*leakage - Eraw*gamma + Eraw - leakage + gamma) /
            (Eraw*leakage*Sraw - Eraw*Sraw*gamma + Eraw*Sraw - leakage*Sraw -
             Sraw*dir_ex_t + Sraw*gamma - Sraw + dir_ex_t + 1))

def uncorrect_S(E_R, S, gamma, L_k, d_dirT):
    """Function used to test :func:`correct_S`."""
    return (S*(d_dirT + 1) /
            (-E_R*L_k*S + E_R*L_k + E_R*S*gamma - E_R*S - E_R*gamma +
             E_R + L_k*S - L_k + S*d_dirT - S*gamma + S + gamma))

def test_fretmath():
    """Run a few consistency checks for the correction functions.
    """
    Ex = np.arange(-0.2, 1.2, 0.01)

    # Tests tautology for gamma = 1
    assert np.allclose(Ex, correct_E_gamma_leak_dir(Ex, gamma=1))
    assert np.allclose(Ex, uncorrect_E_gamma_leak_dir(Ex, gamma=1))
    assert np.allclose(Ex, gamma_correct_E(Ex, gamma=1))
    assert np.allclose(Ex, gamma_uncorrect_E(Ex, gamma=1))

    # Tests tautology for leakage = 0
    assert np.allclose(Ex, leakage_correct_E(Ex, leakage=0))
    assert np.allclose(Ex, leakage_uncorrect_E(Ex, leakage=0))

    # Tests tautology for dir_ex_t = 0
    assert np.allclose(Ex, dir_ex_correct_E(Ex, dir_ex_t=0))
    assert np.allclose(Ex, dir_ex_uncorrect_E(Ex, dir_ex_t=0))

    # Test round-trip consistency
    for leakage in [0.01, 0.04, 0.2]:
        for dir_ex_t in [0.01, 0.03, 0.09]:
            for gamma in [0.2, 0.5, 0.8, 1.2, 1.8]:
                Ec = correct_E_gamma_leak_dir(Ex, gamma, leakage, dir_ex_t)
                Eu = uncorrect_E_gamma_leak_dir(Ec, gamma, leakage, dir_ex_t)
                assert np.allclose(Ex, Eu)

    for gamma in [0.2, 0.5, 0.8, 1.2, 1.8]:
        # Test gamma correction
        E1 = gamma_correct_E(Ex, gamma)
        E2 = correct_E_gamma_leak_dir(Ex, gamma)
        assert np.allclose(E1, E2)
        E1 = gamma_uncorrect_E(Ex, gamma)
        E2 = uncorrect_E_gamma_leak_dir(Ex, gamma)
        assert np.allclose(E1, E2)

    for leakage in [0.01, 0.04, 0.2]:
        # Test leakage correction
        E1 = leakage_correct_E(Ex, leakage)
        E2 = correct_E_gamma_leak_dir(Ex, leakage=leakage)
        assert np.allclose(E1, E2)
        E1 = leakage_uncorrect_E(Ex, leakage)
        E2 = uncorrect_E_gamma_leak_dir(Ex, leakage=leakage)
        assert np.allclose(E1, E2)

    for dir_ex_t in [0.01, 0.03, 0.09]:
        # Test direct excitation correction
        E1 = dir_ex_correct_E(Ex, dir_ex_t)
        E2 = correct_E_gamma_leak_dir(Ex, dir_ex_t=dir_ex_t)
        assert np.allclose(E1, E2)
        E1 = dir_ex_uncorrect_E(Ex, dir_ex_t)
        E2 = uncorrect_E_gamma_leak_dir(Ex, dir_ex_t=dir_ex_t)
        assert np.allclose(E1, E2)

    # Test S correction
    np.random.seed(1)
    Ex = np.arange(-0.2, 1.2, 0.01)
    Sx = np.arange(-0.2, 1.2, 0.01)
    np.random.shuffle(Ex)
    np.random.shuffle(Sx)

    gamma_ = 1
    leakage_ = 0
    dir_ex_t_ = 0
    S_corr = correct_S(Ex, Sx, gamma_, leakage_, dir_ex_t_)
    S_uncorr = uncorrect_S(Ex, S_corr, gamma_, leakage_, dir_ex_t_)
    assert np.allclose(S_corr, Sx)
    assert np.allclose(S_uncorr, Sx)

    gamma_ = 0.7
    leakage_ = 0.05
    dir_ex_t_ = 0.1
    S_corr = correct_S(Ex, Sx, gamma_, leakage_, dir_ex_t_)
    S_uncorr = uncorrect_S(Ex, S_corr, gamma_, leakage_, dir_ex_t_)
    assert np.allclose(S_uncorr, Sx)
    assert (S_corr.min() > -0.5) and (S_corr.max() < 1.5)

if __name__ == '__main__':
    test_fretmath()
    print('All tests passed.')



"""
This folder contains one file per each supported data file.

Each file contains the binary load and preprocessing functions needed for
the specific format. Functions defined here are used by loader functions
in `loaders.py` to properly initialize a `Data()` variable.
"""